/*
  PLAY7 - by John Henckel  first implementation 2/10/93

  This is a program to play polyphonic music on the PC speaker.
  It is written for the Power-C compiler version 1.0.

  This uses a chord generating technique developed by Steve Muenter.
  His routine, called TRI, ran in a BASIC program called POLY.  It's
  available as Public Domain.

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dos.h>
#include <math.h>
#include <time.h>
#include "play.h"

#define chunk 50
#define nv 7
#define nf 192

typedef struct chord {
  long dura;
  short freq[nv];
} chord;

chord *song;
int len;
float sf;               /* speed factor */
int pa;                 /* ftable[pa] is note a1 */
float sp = 1.0;         /* command line speed adjust */
int pt = 0;             /* command line pitch adjust */
int spd,tr;            /* current speed, transpose */
int ky,ka;              /* keysig and keyadjust */
int ln=0;     /* line number */
int md=0;     /* mode, 0=normal, 1=stacatto, 2=legatto */
int hstep[7]={0,2,3,5,7,8,10};   /* half steps for notes a,b,...,g */
int ftable[nf];        /* frequency table, middle C=64 */
char keysig[7][7] = {
 /* a b c d e f g */
   {2,1,2,1,1,2,1},    /* a */     /* 0=flat, 2=sharp */
   {2,1,2,2,1,2,2},    /* b */
   {1,1,1,1,1,1,1},    /* c */
   {1,1,2,1,1,2,1},    /* d */
   {1,1,2,2,1,2,2},    /* e */
   {1,0,1,1,1,1,1},    /* f */
   {1,1,1,1,1,2,1}};   /* g */

readnotes(char *s)
{
  int i,j,n,c,h,p,q;

  for (i=0; i<nv; ++i) {
    while (*s<=' ' && *s) ++s;
    if (!*s) break;                     /* end of string */
    c=*s;
    n=0;
    if (c>='A' && c<='Z') c+='a'-'A';
    if (c>='a' && c<='g') {
      j=1; h=0;
      if (s[j]=='#' || s[j]=='+') { j++; h+=1; }
      if (s[j]=='#' || s[j]=='+') { j++; h+=1; }
      if (s[j]=='b' || s[j]=='-') { j++; h-=1; }
      if (s[j]=='b' || s[j]=='-') { j++; h-=1; }
      q=0;
      if (s[j]>='0' && s[j]<='9')
        sscanf(s+j,"%d",&n);
      else if (i==0) n=3;               /* default octave */
      else q=1;                         /* compute octave based on prev */
      n=n*12+keysig[ky][c-'a']+hstep[c-'a']+ka+h+pa;
      while (q && n<=p) n+=12;
    }
    else if (c>='0' && c<='9') {
      sscanf(s,"%d",&n);
      n+=pa;
    }
    else if (c=='r') ;                  /* do nothing */
    else if (c=='z') ;
    else
      printf("Unknown note '%.2s' on line %d\n",s,ln);

    if (c=='z') {
      j=sscanf(s+1,"%d",&n);
      if (!j) printf("Invalid frequency '%.3s' on line %d\n",s,ln);
      song[len].freq[i] = n;
    }
    else {
      if (n<0 || n>nf) printf("Note '%.2s' out of range on line %d\n",s,ln);
      while (n<0) n+=12;
      while (n>nf) n-=12;
      song[len].freq[i] = ftable[n];
    }
    p=n;
    while (*s>' ') ++s;
  }
  /* fill in the rest of the notes */

  if (i==0) song[len].freq[i++] = 0;
  for (j=i; j<nv; ++j)
    song[len].freq[j] = song[len].freq[j%i];

  return 0;
}

#define growsong \
      if (++len%chunk==0 && \
         (song=realloc(song,(len+chunk)*sizeof*song))==NULL) \
        { printf("Out of memory\n."); exit(1); }

readsong(FILE *f)
{
  static char dat[1000];
  int v,rc;
  int i;

  spd=800;
  tr=0;
  ky=2; ka=0;           /* C major default */
  len=1;                         /* skip first chord (used by asm) */
  song=malloc(chunk*sizeof*song);

  while (1) {
    rc=fscanf(f,"%s",dat);
    if (rc==EOF) break;
    ++ln;
    if (*dat=='s' || *dat=='S') {               /* speed */
      rc=fscanf(f,"%d",&spd);
      if (!rc) printf("Invalid speed on line %d\n",ln);
    }
    else if (*dat=='t' || *dat=='T') {          /* trans */
      rc=fscanf(f,"%d",&tr);
      if (!rc) printf("Invalid transpose on line %d\n",ln);
    }
    else if (*dat=='k' || *dat=='K') {          /* keysig */
      fscanf(f,"%s",dat);
      if (*dat>='A' && *dat<='G') *dat+='a'-'A';
      if (*dat>='a' && *dat<='g') ky=*dat-'a';
      else printf("Invalid keysig %s on line %d\n",dat,ln);
      i=1; ka=0;
      if (dat[i]=='-' || dat[i]=='b') { ++i; ka=-1; }
      if (dat[i]=='+' || dat[i]=='#') { ++i; ka=+1; }
      if (dat[i]=='m') ky=(ky+2)%7;                 /* Minor key */
    }
    else if (*dat=='c' || *dat=='C') {          /* chord */
      rc=fscanf(f,"%d",&v);
      if (!rc) printf("Invalid value (chord duration) on line %d\n",ln);
      if (v<1) v=24;
      song[len].dura = sf/spd/v/sp;
      if (fgetc(f)=='.') { song[len].dura *= 3.0/2;
        if (fgetc(f)=='.') { song[len].dura *= 7.0/6;
          if (fgetc(f)=='.') { song[len].dura *= 15.0/14; }}}
      rc=(dat[1]=='t' || dat[1]=='T');
      fgets(dat,sizeof dat,f);
      readnotes(dat);
      if (!rc && md==0) {                /* mode normal */
        growsong
        memset(song+len,0,sizeof*song);
        if (v>4) i = song[len-1].dura/16;     /* at most a 64th rest */
        else i = v*song[len-1].dura/64;
        song[len-1].dura -= i;
        song[len].dura = i;
      }
      if (!rc && md==1) {                /* mode staccatto */
        growsong
        memset(song+len,0,sizeof*song);
        if (v>4) i = song[len-1].dura/2;     /* at most an 8th note */
        else i = v*song[len-1].dura/8;
        song[len].dura = song[len-1].dura-i;
        song[len-1].dura = i;
      }
      growsong
      song[len].dura = 0;    /* end of song */
    }
    else if (*dat=='m' || *dat=='M') {          /* mode */
      rc=fscanf(f,"%d",&v);
      fgets(dat,sizeof dat,f);
           if (*dat=='n' || *dat=='N') md=0;
      else if (*dat=='s' || *dat=='S') md=1;
      else if (*dat=='l' || *dat=='L') md=2;
      else printf("Unknown mode '%.2s' on line %d\n",dat,ln);
    }
    else if (*dat=='r' || *dat=='R') {          /* repeat */
      rc=fscanf(f,"%d",&v);
      if (!rc || v<1 || v>=len) printf("Invalid repeat on line %d\n",ln);
      else for (i=0; i<v; ++i) {
        memcpy(song+len,song+len-v,sizeof*song);
        growsong
      }
      song[len].dura = 0;    /* end of song */
    }
    else if (*dat=='*' || *dat==';') {          /* comment */
      fgets(dat,sizeof dat,f);
    }
    else if (*dat=='e' || *dat=='E') {          /* echo */
      fgets(dat,sizeof dat,f);
      printf("%s",dat+1);
    }

    else {
      printf("unknown command '%s' ignored.\n",dat);
      fgets(dat,sizeof dat,f);
    }
  }
  return 0;
}

/*
   This sets sf,pa and fills the ftable.
*/
init()
{
  long i;
  clock_t t;

  t=clock();
  i=0;
  while (clock()-t<CLK_TCK) ++i;    /* see how high we can count in a second */
  if (i<=0) i=1;
  sf = i*100000.0;
  pa = nf-10-12*log(i)/log(2);   /* 10 is arbitrary */
  if (pa>nf-89) pa=nf-89;
  if (pa<0) pa=0;

  for (i=0; i<nf; i++) {
    ftable[i] = exp(0.05776227*i)+0.43;
  }
  return 0;
}

dumpsong()
{
  int i,j;
  for (i=0; i<len; ++i) {
    printf("%d. %10ld",i,song[i].dura);
    for (j=0; j<nv; ++j) printf(" %6d",song[i].freq[j]);
    printf("\n");
  }
  return 0;
}

#define p(x) printf(x);

main(int argc,char **argv)
{
  char fn[200];
  FILE *f;

  if (argc<2) {
  p("Piano Song Player                    John Henckel "__DATE__"\n")
  p("This a program to play chords on a PC.  The basic idea is from\n")
  p("Steve Muenters program called TRI or POLY.  This program plays\n")
  p("song files which you can construct using a text editor.  The song\n")
  p("file can have these instructions (one per line):\n")
  p("   speed 800                 set the speed, 800=normal\n")
  p("   tran +5                   transpose, number of half-steps\n")
  p("   chord 4. 39 e2 g#3 z1200  play a chord, up to 7 notes\n")
  p("   repeat 3                  repeat the last 3 chords\n")
  p("   key F#m                   key signature is F-sharp minor (3 sharps)\n")
  p("   mode stacatto             can be stacatto, normal, or legatto\n")
  p("- Use speed 400 to play at half speed.\n")
  p("- Use tran -12 to lower the song one octave.\n")
  p("- The first number after chord is the value of the note, it is\n")
  p("  4 divided by the number of beats.  For example in 4/4 time,\n")
  p("  1=whole note, 2=half, 4=quarter, 4.=dotted quarter, etc.\n")
  p("  You can assign the values differently.  The higher the value\n")
  p("  the shorter the note.\n")
  p("- The notes following the chord indicate the keys on a piano. They\n")
  p("  can be either by letter or by number.  By number, the lowest key A\n")
  p("  is 1 and the highest key C is 88; middle C is 40.  By letter, a1\n")
  p("  (or A1) is the lowest key, and c#4 is c-sharp of the fourth octave\n")
  p("  (next to middle C) and g- is g flat.  The default octave of the\n")
  p("  first note is 3, other notes are (by default) just above the\n")
  p("  preceeding note in the chord.  Use z to play absolute frequencies.\n")
  p("  Use ct instead of 'chord' to tie it to the following chord, this\n")
  p("  is like a temporary legatto. \n")
  p("- All commands can be abbreviated to one letter.\n")
  p("- Use repeat 3 to repeat the last 3 chords one time.  You can use two\n")
  p("  repeats to repeat twice.\n\n")
  p("Command line parameters:  songfile speed-adjust pitch-adjust dump\n")
  p("  The default file extension is .COR.\n");
  p("  Speed and pitch adjust are optional, defaults are 1.0 and 0, resp.");
    return 0;
  }
  strcpy(fn,argv[1]);
  if (strchr(fn,'.')==NULL) strcat(fn,".COR");
  f = fopen(fn,"r");
  if (f==NULL) return printf("unable to read %s",fn);

  init();
  if (argc>2) sscanf(argv[2],"%g",&sp);
  if (argc>3) sscanf(argv[3],"%d",&pt);
  readsong(f);
  fclose(f);
  if (argc>4 && argv[4][0]=='d') {
    printf("Machine speed-factor=%g, pitch adjust=%d\n",sf,pa);
    printf("Song len %d, Cmd line speed %g, pitch %d\n",len,sp,pt);
    printf("Current speed %d, trans %d, key %d+%d, mode %d\n",spd,tr,ky,ka,md);
    printf("Num (nonblank) lines %d\n",ln);
    dumpsong();
  }
  else
    asm(asmcode,song);
}

